package solutions.bitbadger.documents.scala

import org.scalatest.funspec.AnyFunSpec
import org.scalatest.matchers.should.Matchers
import solutions.bitbadger.documents.support.ForceDialect
import solutions.bitbadger.documents.{Dialect, DocumentException, Field, FieldFormat, Op}

import scala.jdk.CollectionConverters.*

class FieldSpec extends AnyFunSpec with ClearConfiguration with Matchers {

  // ~~~ INSTANCE METHODS ~~~

  describe("withParameterName") {
    it("fails for invalid name") {
      a [DocumentException] should be thrownBy Field.equal("it", "").withParameterName("2424")
    }
    it("works with colon prefix") {
      val field = Field.equal("abc", "22").withQualifier("me")
      val withParam = field.withParameterName(":test")
      withParam should not be theSameInstanceAs (field)
      withParam.getName shouldEqual field.getName
      withParam.getComparison shouldEqual field.getComparison
      withParam.getParameterName shouldEqual ":test"
      withParam.getQualifier shouldEqual field.getQualifier
    }
    it("works with at-sign prefix") {
      val field = Field.equal("def", "44")
      val withParam = field.withParameterName("@unit")
      withParam should not be theSameInstanceAs (field)
      withParam.getName shouldEqual field.getName
      withParam.getComparison shouldEqual field.getComparison
      withParam.getParameterName shouldEqual "@unit"
      withParam.getQualifier shouldEqual field.getQualifier
    }
  }

  describe("withQualifier") {
    it("sets qualifier correctly") {
      val field = Field.equal("j", "k")
      val withQual = field.withQualifier("test")
      withQual should not be theSameInstanceAs (field)
      withQual.getName shouldEqual field.getName
      withQual.getComparison shouldEqual field.getComparison
      withQual.getParameterName shouldEqual field.getParameterName
      withQual.getQualifier shouldEqual "test"
    }
  }

  describe("path") {
    it("generates simple unqualified PostgreSQL field") {
      Field.greaterOrEqual("SomethingCool", 18).path(Dialect.POSTGRESQL, FieldFormat.SQL) shouldEqual
        "data->>'SomethingCool'"
    }
    it("generates simple qualified PostgreSQL field") {
      Field.less("SomethingElse", 9).withQualifier("this").path(Dialect.POSTGRESQL, FieldFormat.SQL) shouldEqual
        "this.data->>'SomethingElse'"
    }
    it("generates nested unqualified PostgreSQL field") {
      Field.equal("My.Nested.Field", "howdy").path(Dialect.POSTGRESQL, FieldFormat.SQL) shouldEqual
        "data#>>'{My,Nested,Field}'"
    }
    it("generates nested qualified PostgreSQL field") {
      Field.equal("Nest.Away", "doc").withQualifier("bird").path(Dialect.POSTGRESQL, FieldFormat.SQL) shouldEqual
        "bird.data#>>'{Nest,Away}'"
    }
    it("generates simple unqualified SQLite field") {
      Field.greaterOrEqual("SomethingCool", 18).path(Dialect.SQLITE, FieldFormat.SQL) shouldEqual
        "data->>'SomethingCool'"
    }
    it("generates simple qualified SQLite field") {
      Field.less("SomethingElse", 9).withQualifier("this").path(Dialect.SQLITE, FieldFormat.SQL) shouldEqual
        "this.data->>'SomethingElse'"
    }
    it("generates nested unqualified SQLite field") {
      Field.equal("My.Nested.Field", "howdy").path(Dialect.SQLITE, FieldFormat.SQL) shouldEqual
        "data->'My'->'Nested'->>'Field'"
    }
    it("generates nested qualified SQLite field") {
      Field.equal("Nest.Away", "doc").withQualifier("bird").path(Dialect.SQLITE, FieldFormat.SQL) shouldEqual
        "bird.data->'Nest'->>'Away'"
    }
  }

  describe("toWhere") {
    it("generates exists w/o qualifier | PostgreSQL") {
      ForceDialect.postgres()
      Field.exists("that_field").toWhere shouldEqual "data->>'that_field' IS NOT NULL"
    }
    it("generates exists w/o qualifier | SQLite") {
      ForceDialect.sqlite()
      Field.exists("that_field").toWhere shouldEqual "data->>'that_field' IS NOT NULL"
    }
    it("generates not-exists w/o qualifier | PostgreSQL") {
      ForceDialect.postgres()
      Field.notExists("a_field").toWhere shouldEqual "data->>'a_field' IS NULL"
    }
    it("generates not-exists w/o qualifier | SQLite") {
      ForceDialect.sqlite()
      Field.notExists("a_field").toWhere shouldEqual "data->>'a_field' IS NULL"
    }
    it("generates BETWEEN w/o qualifier, numeric range | PostgreSQL") {
      ForceDialect.postgres()
      Field.between("age", 13, 17, "@age").toWhere shouldEqual "(data->>'age')::numeric BETWEEN @agemin AND @agemax"
    }
    it("generates BETWEEN w/o qualifier, alphanumeric range | PostgreSQL") {
      ForceDialect.postgres()
      Field.between("city", "Atlanta", "Chicago", ":city").toWhere shouldEqual
        "data->>'city' BETWEEN :citymin AND :citymax"
    }
    it("generates BETWEEN w/o qualifier | SQLite") {
      ForceDialect.sqlite()
      Field.between("age", 13, 17, "@age").toWhere shouldEqual "data->>'age' BETWEEN @agemin AND @agemax"
    }
    it("generates BETWEEN w/ qualifier, numeric range | PostgreSQL") {
      ForceDialect.postgres()
      Field.between("age", 13, 17, "@age").withQualifier("test").toWhere shouldEqual
        "(test.data->>'age')::numeric BETWEEN @agemin AND @agemax"
    }
    it("generates BETWEEN w/ qualifier, alphanumeric range | PostgreSQL") {
      ForceDialect.postgres()
      Field.between("city", "Atlanta", "Chicago", ":city").withQualifier("unit").toWhere shouldEqual
        "unit.data->>'city' BETWEEN :citymin AND :citymax"
    }
    it("generates BETWEEN w/ qualifier | SQLite") {
      ForceDialect.sqlite()
      Field.between("age", 13, 17, "@age").withQualifier("my").toWhere shouldEqual
        "my.data->>'age' BETWEEN @agemin AND @agemax"
    }
    it("generates IN/any, numeric values | PostgreSQL") {
      ForceDialect.postgres()
      Field.any("even", List(2, 4, 6).asJava, ":nbr").toWhere shouldEqual
        "(data->>'even')::numeric IN (:nbr_0, :nbr_1, :nbr_2)"
    }
    it("generates IN/any, alphanumeric values | PostgreSQL") {
      ForceDialect.postgres()
      Field.any("test", List("Atlanta", "Chicago").asJava, ":city").toWhere shouldEqual
        "data->>'test' IN (:city_0, :city_1)"
    }
    it("generates IN/any | SQLite") {
      ForceDialect.sqlite()
      Field.any("test", List("Atlanta", "Chicago").asJava, ":city").toWhere shouldEqual
        "data->>'test' IN (:city_0, :city_1)"
    }
    it("generates inArray | PostgreSQL") {
      ForceDialect.postgres()
      Field.inArray("even", "tbl", List(2, 4, 6, 8).asJava, ":it").toWhere shouldEqual
        "data->'even' ??| ARRAY[:it_0, :it_1, :it_2, :it_3]"
    }
    it("generates inArray | SQLite") {
      ForceDialect.sqlite()
      Field.inArray("test", "tbl", List("Atlanta", "Chicago").asJava, ":city").toWhere shouldEqual
        "EXISTS (SELECT 1 FROM json_each(tbl.data, '$.test') WHERE value IN (:city_0, :city_1))"
    }
    it("generates others w/o qualifier | PostgreSQL") {
      ForceDialect.postgres()
      Field.equal("some_field", "", ":value").toWhere shouldEqual "data->>'some_field' = :value"
    }
    it("generates others w/o qualifier | SQLite") {
      ForceDialect.sqlite()
      Field.equal("some_field", "", ":value").toWhere shouldEqual "data->>'some_field' = :value"
    }
    it("generates no-parameter w/ qualifier | PostgreSQL") {
      ForceDialect.postgres()
      Field.exists("no_field").withQualifier("test").toWhere shouldEqual "test.data->>'no_field' IS NOT NULL"
    }
    it("generates no-parameter w/ qualifier | SQLite") {
      ForceDialect.sqlite()
      Field.exists("no_field").withQualifier("test").toWhere shouldEqual "test.data->>'no_field' IS NOT NULL"
    }
    it("generates parameter w/ qualifier | PostgreSQL") {
      ForceDialect.postgres()
      Field.lessOrEqual("le_field", 18, ":it").withQualifier("q").toWhere shouldEqual
        "(q.data->>'le_field')::numeric <= :it"
    }
    it("generates parameter w/ qualifier | SQLite") {
      ForceDialect.sqlite()
      Field.lessOrEqual("le_field", 18, ":it").withQualifier("q").toWhere shouldEqual "q.data->>'le_field' <= :it"
    }
  }

  // ~~~ STATIC CONSTRUCTOR TESTS ~~~

  describe("equal") {
    it("constructs a field w/o parameter name") {
      val field = Field.equal("Test", 14)
      field.getName shouldEqual "Test"
      field.getComparison.getOp shouldEqual Op.EQUAL
      field.getComparison.getValue shouldEqual 14
      field.getParameterName should be (null)
      field.getQualifier should be (null)
    }
    it("constructs a field w/ parameter name") {
      val field = Field.equal("Test", 14, ":w")
      field.getName shouldEqual "Test"
      field.getComparison.getOp shouldEqual Op.EQUAL
      field.getComparison.getValue shouldEqual 14
      field.getParameterName shouldEqual ":w"
      field.getQualifier should be (null)
    }
  }

  describe("greater") {
    it("constructs a field w/o parameter name") {
      val field = Field.greater("Great", "night")
      field.getName shouldEqual "Great"
      field.getComparison.getOp shouldEqual Op.GREATER
      field.getComparison.getValue shouldEqual "night"
      field.getParameterName should be (null)
      field.getQualifier should be (null)
    }
    it("constructs a field w/ parameter name") {
      val field = Field.greater("Great", "night", ":yeah")
      field.getName shouldEqual "Great"
      field.getComparison.getOp shouldEqual Op.GREATER
      field.getComparison.getValue shouldEqual "night"
      field.getParameterName shouldEqual ":yeah"
      field.getQualifier should be (null)
    }
  }

  describe("greaterOrEqual") {
    it("constructs a field w/o parameter name") {
      val field = Field.greaterOrEqual("Nice", 88L)
      field.getName shouldEqual "Nice"
      field.getComparison.getOp shouldEqual Op.GREATER_OR_EQUAL
      field.getComparison.getValue shouldEqual 88L
      field.getParameterName should be (null)
      field.getQualifier should be (null)
    }
    it("constructs a field w/ parameter name") {
      val field = Field.greaterOrEqual("Nice", 88L, ":nice")
      field.getName shouldEqual "Nice"
      field.getComparison.getOp shouldEqual Op.GREATER_OR_EQUAL
      field.getComparison.getValue shouldEqual 88L
      field.getParameterName shouldEqual ":nice"
      field.getQualifier should be (null)
    }
  }

  describe("less") {
    it("constructs a field w/o parameter name") {
      val field = Field.less("Lesser", "seven")
      field.getName shouldEqual "Lesser"
      field.getComparison.getOp shouldEqual Op.LESS
      field.getComparison.getValue shouldEqual "seven"
      field.getParameterName should be (null)
      field.getQualifier should be (null)
    }
    it("constructs a field w/ parameter name") {
      val field = Field.less("Lesser", "seven", ":max")
      field.getName shouldEqual "Lesser"
      field.getComparison.getOp shouldEqual Op.LESS
      field.getComparison.getValue shouldEqual "seven"
      field.getParameterName shouldEqual ":max"
      field.getQualifier should be (null)
    }
  }

  describe("lessOrEqual") {
    it("constructs a field w/o parameter name") {
      val field = Field.lessOrEqual("Nobody", "KNOWS")
      field.getName shouldEqual "Nobody"
      field.getComparison.getOp shouldEqual Op.LESS_OR_EQUAL
      field.getComparison.getValue shouldEqual "KNOWS"
      field.getParameterName should be (null)
      field.getQualifier should be (null)
    }
    it("constructs a field w/ parameter name") {
      val field = Field.lessOrEqual("Nobody", "KNOWS", ":nope")
      field.getName shouldEqual "Nobody"
      field.getComparison.getOp shouldEqual Op.LESS_OR_EQUAL
      field.getComparison.getValue shouldEqual "KNOWS"
      field.getParameterName shouldEqual ":nope"
      field.getQualifier should be (null)
    }
  }

  describe("notEqual") {
    it("constructs a field w/o parameter name") {
      val field = Field.notEqual("Park", "here")
      field.getName shouldEqual "Park"
      field.getComparison.getOp shouldEqual Op.NOT_EQUAL
      field.getComparison.getValue shouldEqual "here"
      field.getParameterName should be (null)
      field.getQualifier should be (null)
    }
    it("constructs a field w/ parameter name") {
      val field = Field.notEqual("Park", "here", ":now")
      field.getName shouldEqual "Park"
      field.getComparison.getOp shouldEqual Op.NOT_EQUAL
      field.getComparison.getValue shouldEqual "here"
      field.getParameterName shouldEqual ":now"
      field.getQualifier should be (null)
    }
  }

  describe("between") {
    it("constructs a field w/o parameter name") {
      val field = Field.between("Age", 18, 49)
      field.getName shouldEqual "Age"
      field.getComparison.getOp shouldEqual Op.BETWEEN
      field.getComparison.getValue.getFirst shouldEqual 18
      field.getComparison.getValue.getSecond shouldEqual 49
      field.getParameterName should be (null)
      field.getQualifier should be (null)
    }
    it("constructs a field w/ parameter name") {
      val field = Field.between("Age", 18, 49, ":limit")
      field.getName shouldEqual "Age"
      field.getComparison.getOp shouldEqual Op.BETWEEN
      field.getComparison.getValue.getFirst shouldEqual 18
      field.getComparison.getValue.getSecond shouldEqual 49
      field.getParameterName shouldEqual ":limit"
      field.getQualifier should be (null)
    }
  }

  describe("any") {
    it("constructs a field w/o parameter name") {
      val field = Field.any("Here", List(8, 16, 32).asJava)
      field.getName shouldEqual "Here"
      field.getComparison.getOp shouldEqual Op.IN
      field.getComparison.getValue should be (List(8, 16, 32).asJava)
      field.getParameterName should be (null)
      field.getQualifier should be (null)
    }
    it("constructs a field w/ parameter name") {
      val field = Field.any("Here", List(8, 16, 32).asJava, ":list")
      field.getName shouldEqual "Here"
      field.getComparison.getOp shouldEqual Op.IN
      field.getComparison.getValue should be (List(8, 16, 32).asJava)
      field.getParameterName shouldEqual ":list"
      field.getQualifier should be (null)
    }
  }

  describe("inArray") {
    it("constructs a field w/o parameter name") {
      val field = Field.inArray("ArrayField", "table", List("z").asJava)
      field.getName shouldEqual "ArrayField"
      field.getComparison.getOp shouldEqual Op.IN_ARRAY
      field.getComparison.getValue.getFirst shouldEqual "table"
      field.getComparison.getValue.getSecond should be (List("z").asJava)
      field.getParameterName should be (null)
      field.getQualifier should be (null)
    }
    it("constructs a field w/ parameter name") {
      val field = Field.inArray("ArrayField", "table", List("z").asJava, ":a")
      field.getName shouldEqual "ArrayField"
      field.getComparison.getOp shouldEqual Op.IN_ARRAY
      field.getComparison.getValue.getFirst shouldEqual "table"
      field.getComparison.getValue.getSecond should be (List("z").asJava)
      field.getParameterName shouldEqual ":a"
      field.getQualifier should be (null)
    }
  }

  describe("exists") {
    it("constructs a field") {
      val field = Field.exists("Groovy")
      field.getName shouldEqual "Groovy"
      field.getComparison.getOp shouldEqual Op.EXISTS
      field.getComparison.getValue shouldEqual ""
      field.getParameterName should be (null)
      field.getQualifier should be (null)
    }
  }

  describe("notExists") {
    it("constructs a field") {
      val field = Field.notExists("Groovy")
      field.getName shouldEqual "Groovy"
      field.getComparison.getOp shouldEqual Op.NOT_EXISTS
      field.getComparison.getValue shouldEqual ""
      field.getParameterName should be (null)
      field.getQualifier should be (null)
    }
  }

  describe("named") {
    it("named constructs a field") {
      val field = Field.named("Tacos")
      field.getName shouldEqual "Tacos"
      field.getComparison.getOp shouldEqual Op.EQUAL
      field.getComparison.getValue shouldEqual ""
      field.getParameterName should be (null)
      field.getQualifier should be (null)
    }
  }

  describe("static constructors") {
    it("fail for invalid parameter name") {
      a [DocumentException] should be thrownBy Field.equal("a", "b", "that ain't it, Jack...")
    }
  }

  describe("nameToPath") {
    it("creates a simple PostgreSQL SQL name") {
      Field.nameToPath("Simple", Dialect.POSTGRESQL, FieldFormat.SQL) shouldEqual "data->>'Simple'"
    }
    it("creates a simple SQLite SQL name") {
      Field.nameToPath("Simple", Dialect.SQLITE, FieldFormat.SQL) shouldEqual "data->>'Simple'"
    }
    it("creates a nested PostgreSQL SQL name") {
      Field.nameToPath("A.Long.Path.to.the.Property", Dialect.POSTGRESQL, FieldFormat.SQL) shouldEqual
        "data#>>'{A,Long,Path,to,the,Property}'"
    }
    it("creates a nested SQLite SQL name") {
      Field.nameToPath("A.Long.Path.to.the.Property", Dialect.SQLITE, FieldFormat.SQL) shouldEqual
        "data->'A'->'Long'->'Path'->'to'->'the'->>'Property'"
    }
    it("creates a simple PostgreSQL JSON name") {
      Field.nameToPath("Simple", Dialect.POSTGRESQL, FieldFormat.JSON) shouldEqual "data->'Simple'"
    }
    it("creates a simple SQLite JSON name") {
      Field.nameToPath("Simple", Dialect.SQLITE, FieldFormat.JSON) shouldEqual "data->'Simple'"
    }
    it("creates a nested PostgreSQL JSON name") {
      Field.nameToPath("A.Long.Path.to.the.Property", Dialect.POSTGRESQL, FieldFormat.JSON) shouldEqual
        "data#>'{A,Long,Path,to,the,Property}'"
    }
    it("creates a nested SQLite JSON name") {
      Field.nameToPath("A.Long.Path.to.the.Property", Dialect.SQLITE, FieldFormat.JSON) shouldEqual
        "data->'A'->'Long'->'Path'->'to'->'the'->'Property'"
    }
  }
}
