package solutions.bitbadger.documents.scala.query

import org.scalatest.funspec.AnyFunSpec
import org.scalatest.matchers.should.Matchers
import solutions.bitbadger.documents.{Dialect, DocumentException, DocumentIndex}
import solutions.bitbadger.documents.query.DefinitionQuery
import solutions.bitbadger.documents.scala.ClearConfiguration
import solutions.bitbadger.documents.support.ForceDialect
import solutions.bitbadger.documents.support.TypesKt.TEST_TABLE

import scala.jdk.CollectionConverters.*

class DefinitionQuerySpec extends AnyFunSpec with ClearConfiguration with Matchers {

  describe("ensureTableFor") {
    it("generates correctly") {
      DefinitionQuery.ensureTableFor("my.table", "JSONB") shouldEqual
        "CREATE TABLE IF NOT EXISTS my.table (data JSONB NOT NULL)"
    }
  }

  describe("ensureTable") {
    it("generates correctly | PostgreSQL") {
      ForceDialect.postgres()
      DefinitionQuery.ensureTable(TEST_TABLE) shouldEqual
        s"CREATE TABLE IF NOT EXISTS $TEST_TABLE (data JSONB NOT NULL)"
    }
    it("generates correctly | SQLite") {
      ForceDialect.sqlite()
      DefinitionQuery.ensureTable(TEST_TABLE) shouldEqual
        s"CREATE TABLE IF NOT EXISTS $TEST_TABLE (data TEXT NOT NULL)"
    }
    it("fails when no dialect is set") {
      a [DocumentException] should be thrownBy DefinitionQuery.ensureTable(TEST_TABLE)
    }
  }

  describe("ensureKey") {
    it("generates correctly with schema") {
      DefinitionQuery.ensureKey("test.table", Dialect.POSTGRESQL) shouldEqual
        "CREATE UNIQUE INDEX IF NOT EXISTS idx_table_key ON test.table ((data->>'id'))"
    }
    it("generates correctly without schema") {
      DefinitionQuery.ensureKey(TEST_TABLE, Dialect.SQLITE) shouldEqual
        s"CREATE UNIQUE INDEX IF NOT EXISTS idx_${TEST_TABLE}_key ON $TEST_TABLE ((data->>'id'))"
    }
  }

  describe("ensureIndexOn") {
    it("generates multiple fields and directions") {
      DefinitionQuery.ensureIndexOn("test.table", "gibberish", List("taco", "guac DESC", "salsa ASC").asJava,
          Dialect.POSTGRESQL) shouldEqual
        "CREATE INDEX IF NOT EXISTS idx_table_gibberish ON test.table ((data->>'taco'), (data->>'guac') DESC, (data->>'salsa') ASC)"
    }
    it("generates nested field | PostgreSQL") {
      DefinitionQuery.ensureIndexOn(TEST_TABLE, "nest", List("a.b.c").asJava, Dialect.POSTGRESQL) shouldEqual
        s"CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_nest ON $TEST_TABLE ((data#>>'{a,b,c}'))"
    }
    it("generates nested field | SQLite") {
      DefinitionQuery.ensureIndexOn(TEST_TABLE, "nest", List("a.b.c").asJava, Dialect.SQLITE) shouldEqual
        s"CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_nest ON $TEST_TABLE ((data->'a'->'b'->>'c'))"
    }
  }

  describe("ensureDocumentOn") {
    it("generates Full | PostgreSQL") {
      ForceDialect.postgres()
      DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.FULL) shouldEqual
        s"CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_document ON $TEST_TABLE USING GIN (data)"
    }
    it("generates Optimized | PostgreSQL") {
      ForceDialect.postgres()
      DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.OPTIMIZED) shouldEqual
        s"CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_document ON $TEST_TABLE USING GIN (data jsonb_path_ops)"
    }
    it("fails | SQLite") {
      ForceDialect.sqlite()
      a [DocumentException] should be thrownBy DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.FULL)
    }
  }
}
