package solutions.bitbadger.documents.scala.query

import org.scalatest.funspec.AnyFunSpec
import org.scalatest.matchers.should.Matchers
import solutions.bitbadger.documents.{AutoId, Configuration, DocumentException}
import solutions.bitbadger.documents.query.DocumentQuery
import solutions.bitbadger.documents.scala.ClearConfiguration
import solutions.bitbadger.documents.support.ForceDialect
import solutions.bitbadger.documents.support.TypesKt.TEST_TABLE

class DocumentQuerySpec extends AnyFunSpec with ClearConfiguration with Matchers {

  describe("insert") {
    it("generates no auto ID | PostgreSQL") {
      ForceDialect.postgres()
      DocumentQuery.insert(TEST_TABLE) shouldEqual s"INSERT INTO $TEST_TABLE VALUES (:data)"
    }
    it("generates no auto ID | SQLite") {
      ForceDialect.sqlite()
      DocumentQuery.insert(TEST_TABLE) shouldEqual s"INSERT INTO $TEST_TABLE VALUES (:data)"
    }
    it("generates auto number | PostgreSQL") {
      ForceDialect.postgres()
      DocumentQuery.insert(TEST_TABLE, AutoId.NUMBER) shouldEqual
        s"INSERT INTO $TEST_TABLE VALUES (:data::jsonb || ('{\"id\":' " +
          s"|| (SELECT COALESCE(MAX((data->>'id')::numeric), 0) + 1 FROM $TEST_TABLE) || '}')::jsonb)"
    }
    it("generates auto number | SQLite") {
      ForceDialect.sqlite()
      DocumentQuery.insert(TEST_TABLE, AutoId.NUMBER) shouldEqual
        s"INSERT INTO $TEST_TABLE VALUES (json_set(:data, '$$.id', " +
          s"(SELECT coalesce(max(data->>'id'), 0) + 1 FROM $TEST_TABLE)))"
    }
    it("generates auto UUID | PostgreSQL") {
      ForceDialect.postgres()
      val query = DocumentQuery.insert(TEST_TABLE, AutoId.UUID)
      query should startWith (s"INSERT INTO $TEST_TABLE VALUES (:data::jsonb || '{\"id\":\"")
      query should endWith ("\"}')")
    }
    it("generates auto UUID | SQLite") {
      ForceDialect.sqlite()
      val query = DocumentQuery.insert(TEST_TABLE, AutoId.UUID)
      query should startWith (s"INSERT INTO $TEST_TABLE VALUES (json_set(:data, '$$.id', '")
      query should endWith ("'))")
    }
    it("generates auto random string | PostgreSQL") {
      try {
        ForceDialect.postgres()
        Configuration.idStringLength = 8
        val query = DocumentQuery.insert(TEST_TABLE, AutoId.RANDOM_STRING)
        query should startWith (s"INSERT INTO $TEST_TABLE VALUES (:data::jsonb || '{\"id\":\"")
        query should endWith ("\"}')")
        query.replace(s"INSERT INTO $TEST_TABLE VALUES (:data::jsonb || '{\"id\":\"", "").replace("\"}')", "")
          .length shouldEqual 8
      } finally {
        Configuration.idStringLength = 16
      }
    }
    it("insert generates auto random string | SQLite") {
      ForceDialect.sqlite()
      val query = DocumentQuery.insert(TEST_TABLE, AutoId.RANDOM_STRING)
      query should startWith (s"INSERT INTO $TEST_TABLE VALUES (json_set(:data, '$$.id', '")
      query should endWith ("'))")
      query.replace(s"INSERT INTO $TEST_TABLE VALUES (json_set(:data, '$$.id', '", "").replace("'))", "")
        .length shouldEqual Configuration.idStringLength
    }
    it("fails when no dialect is set") {
      a [DocumentException] should be thrownBy DocumentQuery.insert(TEST_TABLE)
    }
  }

  describe("save") {
    it("generates correctly") {
      ForceDialect.postgres()
      DocumentQuery.save(TEST_TABLE) shouldEqual
        s"INSERT INTO $TEST_TABLE VALUES (:data) ON CONFLICT ((data->>'id')) DO UPDATE SET data = EXCLUDED.data"
    }
  }

  describe("update") {
    it("generates successfully") {
      DocumentQuery.update(TEST_TABLE) shouldEqual s"UPDATE $TEST_TABLE SET data = :data"
    }
  }
}
