package solutions.bitbadger.documents.scala.query

import org.scalatest.funspec.AnyFunSpec
import org.scalatest.matchers.should.Matchers
import solutions.bitbadger.documents.{Dialect, Field, FieldMatch}
import solutions.bitbadger.documents.query.QueryUtils
import solutions.bitbadger.documents.scala.ClearConfiguration
import solutions.bitbadger.documents.support.ForceDialect

import scala.jdk.CollectionConverters.*

class QueryUtilsSpec extends AnyFunSpec with ClearConfiguration with Matchers {

  describe("statementWhere") {
    it("generates correctly") {
      QueryUtils.statementWhere("x", "y") shouldEqual "x WHERE y"
    }
  }

  describe("byId") {
    it("generates a numeric ID query | PostgreSQL") {
      ForceDialect.postgres()
      QueryUtils.byId("test", 9) shouldEqual "test WHERE (data->>'id')::numeric = :id"
    }
    it("generates an alphanumeric ID query | PostgreSQL") {
      ForceDialect.postgres()
      QueryUtils.byId("unit", "18") shouldEqual "unit WHERE data->>'id' = :id"
    }
    it("generates ID query | SQLite") {
      ForceDialect.sqlite()
      QueryUtils.byId("yo", 27) shouldEqual "yo WHERE data->>'id' = :id"
    }
  }

  describe("byFields") {
    it("generates default field query | PostgreSQL") {
      ForceDialect.postgres()
      QueryUtils.byFields("this",
          List(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")).asJava) shouldEqual
        "this WHERE data->>'a' = :the_a AND (data->>'b')::numeric = :b_value"
    }
    it("generates default field query | SQLite") {
      ForceDialect.sqlite()
      QueryUtils.byFields("this",
          List(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")).asJava) shouldEqual
        "this WHERE data->>'a' = :the_a AND data->>'b' = :b_value"
    }
    it("generates ANY field query | PostgreSQL") {
      ForceDialect.postgres()
      QueryUtils.byFields("that", List(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")).asJava,
          FieldMatch.ANY) shouldEqual
        "that WHERE data->>'a' = :the_a OR (data->>'b')::numeric = :b_value"
    }
    it("generates ANY field query | SQLite") {
      ForceDialect.sqlite()
      QueryUtils.byFields("that", List(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")).asJava,
          FieldMatch.ANY) shouldEqual
        "that WHERE data->>'a' = :the_a OR data->>'b' = :b_value"
    }
  }

  describe("orderBy") {
    it("generates for no fields") {
      QueryUtils.orderBy(List().asJava, Dialect.POSTGRESQL) shouldEqual ""
      QueryUtils.orderBy(List().asJava, Dialect.SQLITE) shouldEqual ""
    }
    it("generates single, no direction | PostgreSQL") {
      QueryUtils.orderBy(List(Field.named("TestField")).asJava, Dialect.POSTGRESQL) shouldEqual
        " ORDER BY data->>'TestField'"
    }
    it("generates single, no direction | SQLite") {
      QueryUtils.orderBy(List(Field.named("TestField")).asJava, Dialect.SQLITE) shouldEqual
        " ORDER BY data->>'TestField'"
    }
    it("generates multiple with direction | PostgreSQL") {
      QueryUtils.orderBy(
          List(Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")).asJava,
          Dialect.POSTGRESQL) shouldEqual
        " ORDER BY data#>>'{Nested,Test,Field}' DESC, data->>'AnotherField', data->>'It' DESC"
    }
    it("generates multiple with direction | SQLite") {
      QueryUtils.orderBy(
          List(Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")).asJava,
          Dialect.SQLITE) shouldEqual
        " ORDER BY data->'Nested'->'Test'->>'Field' DESC, data->>'AnotherField', data->>'It' DESC"
    }
    it("generates numeric ordering | PostgreSQL") {
      QueryUtils.orderBy(List(Field.named("n:Test")).asJava, Dialect.POSTGRESQL) shouldEqual
        " ORDER BY (data->>'Test')::numeric"
    }
    it("generates numeric ordering | SQLite") {
      QueryUtils.orderBy(List(Field.named("n:Test")).asJava, Dialect.SQLITE) shouldEqual " ORDER BY data->>'Test'"
    }
    it("generates case-insensitive ordering | PostgreSQL") {
      QueryUtils.orderBy(List(Field.named("i:Test.Field DESC NULLS FIRST")).asJava, Dialect.POSTGRESQL) shouldEqual
        " ORDER BY LOWER(data#>>'{Test,Field}') DESC NULLS FIRST"
    }
    it("generates case-insensitive ordering | SQLite") {
      QueryUtils.orderBy(List(Field.named("i:Test.Field ASC NULLS LAST")).asJava, Dialect.SQLITE) shouldEqual
        " ORDER BY data->'Test'->>'Field' COLLATE NOCASE ASC NULLS LAST"
    }
  }
}
